/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "ap_provider.h"
#include "httpd.h"
#include "http_config.h"
#include "ap_provider.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_ldap.h"

#include "mod_auth.h"

#include "apr_strings.h"
#include "apr_xlate.h"
#define APR_WANT_STRFUNC
#include "apr_want.h"
#include "apr_lib.h"
#include "apu_version.h"

#include <ctype.h>

#if !APR_HAS_LDAP
#error mod_meta_ldap requires APR-util to have LDAP support built in. To fix add --with-ldap to ./configure.
#endif

#define META_LDAP_PREFIX "LDAP_"
#define META_LDAP_FILTER "LDAP"

typedef struct {
    char *host;                     /* Name of the LDAP server (or space separated list) */
    int port;                       /* Port of the LDAP server */
    char *basedn;                   /* Base DN to do all searches from */
    const char *attribute;          /* Attribute to search for */
    const char **attributes;        /* Array of all the attributes to return */
    int scope;                      /* Scope of the search */
    char *filter;                   /* Filter to further limit the search  */
    int secure;                     /* True if SSL connections are requested */
} meta_ldap_parsed_url_t;

typedef struct {
    apr_pool_t *pool;               /* Pool that this config is allocated from */

    /* These parameters are all derived from the MetaLDAPURL directive */
    char *url;                      /* String representation of the URL */
    meta_ldap_parsed_url_t parsed_url;
 
    ap_expr_info_t *key;            /* Expression that evaluates to the key */
    ap_expr_info_t *filter;         /* Expression that evaluates to the filter */
    apr_ldap_deref_e deref;         /* how to handle alias dereferening */
    const char *binddn;             /* DN to bind to server (can be NULL) */
    const char *bindpw;             /* Password to bind to server (can be NULL) */
    const char *meta_prefix;        /* Prefix for environment variables added during authz */

    unsigned int url_set:1;
    unsigned int key_set:1;
    unsigned int filter_set:1;
    unsigned int deref_set:1;
    unsigned int binddn_set:1;
    unsigned int bindpw_set:1;
    unsigned int meta_prefix_set:1;
} meta_ldap_config_t;

module AP_MODULE_DECLARE_DATA meta_ldap_module;

static APR_OPTIONAL_FN_TYPE(uldap_connection_close) *util_ldap_connection_close;
static APR_OPTIONAL_FN_TYPE(uldap_connection_find_uri) *util_ldap_connection_find_uri;
static APR_OPTIONAL_FN_TYPE(uldap_cache_comparedn) *util_ldap_cache_comparedn;
static APR_OPTIONAL_FN_TYPE(uldap_cache_compare) *util_ldap_cache_compare;
static APR_OPTIONAL_FN_TYPE(uldap_cache_check_subgroups) *util_ldap_cache_check_subgroups;
static APR_OPTIONAL_FN_TYPE(uldap_cache_checkuserid) *util_ldap_cache_checkuserid;
static APR_OPTIONAL_FN_TYPE(uldap_cache_getuserdn) *util_ldap_cache_getuserdn;

static apr_hash_t *charset_conversions = NULL;
static char *to_charset = NULL;           /* UTF-8 identifier derived from the charset.conv file */

/* Derive a code page ID give a language name or ID */
static char* derive_codepage_from_lang (apr_pool_t *p, char *language)
{
    char *charset;

    if (!language)          /* our default codepage */
        return apr_pstrdup(p, "ISO-8859-1");

    charset = (char*) apr_hash_get(charset_conversions, language, APR_HASH_KEY_STRING);

    /*
     * Test if language values like 'en-US' return a match from the charset
     * conversion map when shortened to 'en'.
     */
    if (!charset && strlen(language) > 3 && language[2] == '-') {
        char *language_short = apr_pstrndup(p, language, 2);
        charset = (char*) apr_hash_get(charset_conversions, language_short, APR_HASH_KEY_STRING);
    }

    if (charset) {
        charset = apr_pstrdup(p, charset);
    }

    return charset;
}

static apr_xlate_t* get_conv_set (request_rec *r)
{
    char *lang_line = (char*)apr_table_get(r->headers_in, "accept-language");
    char *lang;
    apr_xlate_t *convset;

    if (lang_line) {
        lang_line = apr_pstrdup(r->pool, lang_line);
        for (lang = lang_line;*lang;lang++) {
            if ((*lang == ',') || (*lang == ';')) {
                *lang = '\0';
                break;
            }
        }
        lang = derive_codepage_from_lang(r->pool, lang_line);

        if (lang && (apr_xlate_open(&convset, to_charset, lang, r->pool) == APR_SUCCESS)) {
            return convset;
        }
    }

    return NULL;
}


/*
 * Build the search filter, or at least as much of the search filter that
 * will fit in the buffer, and return APR_EGENERAL if it won't fit, otherwise
 * APR_SUCCESS.
 *
 * The search filter consists of the filter provided with the URL,
 * combined with a filter made up of the attribute provided with the URL,
 * and the actual username passed by the HTTP client. For example, assume
 * that the LDAP URL is
 *
 *   ldap://ldap.airius.com/ou=People, o=Airius?uid??(posixid=*)
 *
 * Further, assume that the userid passed by the client was `userj'.  The
 * search filter will be (&(posixid=*)(uid=userj)).
 */
#define FILTER_LENGTH MAX_STRING_LEN
static apr_status_t meta_ldap_build_filter(char filtbuf[FILTER_LENGTH],
                             request_rec *r,
                             const char *key,
                             const char *filter,
                             meta_ldap_config_t *sec)
{
    char *q;
    const char *p, *filtbuf_end;
    apr_xlate_t *convset = NULL;
    apr_size_t inbytes;
    apr_size_t outbytes;
    char *outbuf;
    int nofilter = 0, len;
    apr_status_t rv = APR_SUCCESS;

    if (!filter) {
        filter = sec->parsed_url.filter;
    }

    if (charset_conversions) {
        convset = get_conv_set(r);
    }

    if (convset) {
        inbytes = strlen(key);
        outbytes = (inbytes+1)*3;
        outbuf = apr_pcalloc(r->pool, outbytes);

        /* Convert the user name to UTF-8.  This is only valid for LDAP v3 */
        if (apr_xlate_conv_buffer(convset, key, &inbytes, outbuf, &outbytes) == APR_SUCCESS) {
            key = outbuf;
        }
    }

    /*
     * No key? we do filter only.
     */
    if (!key) {
        len = apr_snprintf(filtbuf, FILTER_LENGTH, "(%s)", filter);
        if (len >= FILTER_LENGTH - 1) {
            return APR_EINVAL;
        }
        return rv;
    }

    /*
     * Create the first part of the filter, which consists of the
     * config-supplied portions.
     */

    if ((nofilter = (!filter || !*filter || !strcasecmp(filter, "none")))) { 
        len = apr_snprintf(filtbuf, FILTER_LENGTH, "(%s=", sec->parsed_url.attribute);
    }
    else { 
        len = apr_snprintf(filtbuf, FILTER_LENGTH, "(&(%s)(%s=", filter, sec->parsed_url.attribute);
    }

    /*
     * Now add the client-supplied username to the filter, ensuring that any
     * LDAP filter metachars are escaped.
     */
    filtbuf_end = filtbuf + FILTER_LENGTH - 1;
    for (p = key, q = filtbuf + len; *p; ) {
        if (strchr("*()\\", *p) != NULL) {
#if APR_HAS_MICROSOFT_LDAPSDK
            if (q + 3 >= filtbuf_end) { /* accounts for final \0 */
                rv = APR_EGENERAL;
                goto out;
            }
            *q++ = '\\';
            switch ( *p++ )
            {
              case '*':
                *q++ = '2';
                *q++ = 'a';
                break;
              case '(':
                *q++ = '2';
                *q++ = '8';
                break;
              case ')':
                *q++ = '2';
                *q++ = '9';
                break;
              case '\\':
                *q++ = '5';
                *q++ = 'c';
                break;
            }
#else
            if (q + 2 >= filtbuf_end) { /* accounts for final \0 */
                rv = APR_EGENERAL;
                goto out;
            }
            *q++ = '\\';
            *q++ = *p++;
#endif
        }
        else {
            if (q + 1 >= filtbuf_end) { /* accounts for final \0 */
                rv = APR_EGENERAL;
                goto out;
            }
            *q++ = *p++;
        }
    }

    /*
     * Append the closing parens of the filter, unless doing so would
     * overrun the buffer.
     */

    if (nofilter) { 
        if (q + 1 >= filtbuf_end) { /* accounts for final \0 */
            rv = APR_EGENERAL;
            goto out;
        }
        *q++ = ')';
    } 
    else { 
        if (q + 2 >= filtbuf_end) { /* accounts for final \0 */
            rv = APR_EGENERAL;
            goto out;
        }
        *q++ = ')';
        *q++ = ')';
    }

out:
    *q = '\0';
    return rv;
}

static void *create_meta_ldap_dir_config(apr_pool_t *p, char *d)
{
    meta_ldap_config_t *sec =
        (meta_ldap_config_t *)apr_pcalloc(p, sizeof(meta_ldap_config_t));

    sec->pool = p;

    sec->binddn = NULL;
    sec->bindpw = NULL;
    sec->deref = APR_LDAP_DEREF_ALWAYS;

    sec->meta_prefix = META_LDAP_PREFIX;

    return sec;
}

static void *merge_meta_ldap_dir_config(apr_pool_t *p, void *basev,
                                        void *overridesv)
{
    meta_ldap_config_t *dc = apr_pcalloc(p, sizeof(meta_ldap_config_t));
    meta_ldap_config_t *base = (meta_ldap_config_t *) basev;
    meta_ldap_config_t *overrides = (meta_ldap_config_t *) overridesv;

    dc->url = (overrides->url_set == 0) ? base->url :
                        overrides->url;
    dc->parsed_url = (overrides->url_set == 0) ? base->parsed_url :
                        overrides->parsed_url;
    dc->url_set = overrides->url_set || base->url_set;

    dc->key = (overrides->key_set == 0) ? base->key :
                        overrides->key;
    dc->key_set = overrides->key_set || base->key_set;

    dc->filter = (overrides->filter_set == 0) ? base->filter :
                        overrides->filter;
    dc->filter_set = overrides->filter_set || base->filter_set;

    dc->deref = (overrides->deref_set == 0) ? base->deref :
                        overrides->deref;
    dc->deref_set = overrides->deref_set || base->deref_set;

    dc->binddn = (overrides->binddn_set == 0) ? base->binddn :
                        overrides->binddn;
    dc->binddn_set = overrides->binddn_set || base->binddn_set;

    dc->bindpw = (overrides->bindpw_set == 0) ? base->bindpw :
                        overrides->bindpw;
    dc->bindpw_set = overrides->bindpw_set || base->bindpw_set;

    dc->meta_prefix = (overrides->meta_prefix_set == 0) ? base->meta_prefix :
                        overrides->meta_prefix;
    dc->meta_prefix_set = overrides->meta_prefix_set || base->meta_prefix_set;

    return dc;
}

static void set_request_vars(request_rec *r, apr_array_header_t **vals)
{
    const char *prefix = NULL;
    int prefix_len;
    meta_ldap_config_t *sec =
        (meta_ldap_config_t *)ap_get_module_config(r->per_dir_config, &meta_ldap_module);

    prefix = sec->meta_prefix;
    prefix_len = strlen(prefix);

    if (sec->parsed_url.attributes && vals) {
        apr_table_t *e = r->subprocess_env;
        int i = 0;
        while (sec->parsed_url.attributes[i] && vals[i]) {
            char *val;
            char *str = apr_pstrcat(r->pool, prefix, sec->parsed_url.attributes[i], NULL);
            int j = prefix_len;
            while (str[j]) {
                str[j] = apr_toupper(str[j]);
                j++;
            }

            val = apr_buffer_pstrncat(r->pool,
                                      (const apr_buffer_t *)vals[i]->elts,
                                      vals[i]->nelts, "; ", APR_BUFFER_BASE64, NULL);

            apr_table_setn(e, str, val);
            i++;
        }
    }
    return;
}

/*
 * Use the ldap url parsing routines to break up the ldap url into
 * host and port.
 */
static const char *meta_ldap_parse_url(cmd_parms *cmd,
                                       void *config,
                                       const char *url,
                                       const char *mode)
{
    int rc, i;
    apr_ldap_url_desc_t *urld;
#if (APU_MAJOR_VERSION == 1 && APU_MINOR_VERSION >= 7)
    apr_ldap_err_t *result;
#else
    apu_err_t *result;
#endif
    const char *end = url;

    meta_ldap_config_t *sec = config;

    rc = apr_ldap_url_parse(cmd->pool, url, &(urld), &(result));
    if (rc != APR_SUCCESS) {
        return result->reason;
    }

    /* isolate the host/port part of the URL */
    for (i = 0; end && i < 3; i++) {
        end = strchr((char *)end + 1, '/');
    }
    if (end) {
        sec->url = apr_pstrndup(cmd->pool, url, end - url);
    }
    else {
        sec->url = apr_pstrdup(cmd->pool, url);
    }

    /* Set all the values, or at least some sane defaults */
    if (sec->parsed_url.host) {
        sec->parsed_url.host = apr_pstrcat(cmd->pool, urld->lud_host, " ", sec->parsed_url.host, NULL);
    }
    else {
        sec->parsed_url.host = urld->lud_host? apr_pstrdup(cmd->pool, urld->lud_host) : "localhost";
    }
    sec->parsed_url.basedn = urld->lud_dn? apr_pstrdup(cmd->pool, urld->lud_dn) : "";
    if (urld->lud_attrs && urld->lud_attrs[0]) {
        char **attributes;
        int i = 1;
        while (urld->lud_attrs[i]) {
            i++;
        }
        attributes = apr_pcalloc(cmd->pool, sizeof(char *) * (i+1));
        i = 0;
        while (urld->lud_attrs[i]) {
            attributes[i] = apr_pstrdup(cmd->pool, urld->lud_attrs[i]);
            i++;
        }
        sec->parsed_url.attributes = (const char **)attributes;
        sec->parsed_url.attribute = sec->parsed_url.attributes[0];
    }
    else {
        sec->parsed_url.attribute = "uid";
    }

    sec->parsed_url.scope = urld->lud_scope == APR_LDAP_SCOPE_ONELEVEL ?
        APR_LDAP_SCOPE_ONELEVEL : APR_LDAP_SCOPE_SUBTREE;

    if (urld->lud_filter) {
        if (urld->lud_filter[0] == '(') {
            /*
             * Get rid of the surrounding parens; later on when generating the
             * filter, they'll be put back.
             */
            sec->parsed_url.filter = apr_pstrmemdup(cmd->pool, urld->lud_filter+1,
                                                    strlen(urld->lud_filter)-2);
        }
        else {
            sec->parsed_url.filter = apr_pstrdup(cmd->pool, urld->lud_filter);
        }
    }
    else {
        sec->parsed_url.filter = "objectclass=*";
    }

    if (mode) {
        if (0 == strcasecmp("NONE", mode)) {
            sec->parsed_url.secure = APR_LDAP_TLS_NONE;
        }
        else if (0 == strcasecmp("SSL", mode)) {
            sec->parsed_url.secure = APR_LDAP_TLS_SSL;
        }
        else if (0 == strcasecmp("TLS", mode) || 0 == strcasecmp("STARTTLS", mode)) {
            sec->parsed_url.secure = APR_LDAP_TLS_STARTTLS;
        }
        else {
            return "Invalid LDAP connection mode setting: must be one of NONE, "
                   "SSL, or TLS/STARTTLS";
        }
    }

    /* "ldaps" indicates secure ldap connections desired */
    if (strncasecmp(url, "ldaps", 5) == 0)
    {
        sec->parsed_url.secure = APR_LDAP_TLS_SSL;
        sec->parsed_url.port = urld->lud_port? urld->lud_port : APR_LDAPS_PORT;
    }
    else
    {
        sec->parsed_url.port = urld->lud_port? urld->lud_port : APR_LDAP_PORT;
    }

    sec->url_set = 1;

    ap_log_error(APLOG_MARK, APLOG_TRACE1, 0, cmd->server,
                 "auth_ldap url parse: `%s', Url: %s, Host: %s, Port: %d, DN: %s, "
                 "attrib: %s, scope: %s, filter: %s, connection mode: %s",
                 url,
                 sec->url,
                 urld->lud_host,
                 urld->lud_port,
                 urld->lud_dn,
                 urld->lud_attrs? urld->lud_attrs[0] : "(null)",
                 (urld->lud_scope == APR_LDAP_SCOPE_SUBTREE ? "subtree" :
                  urld->lud_scope == APR_LDAP_SCOPE_BASE ? "base" :
                  urld->lud_scope == APR_LDAP_SCOPE_ONELEVEL ? "onelevel" :
                  urld->lud_scope == APR_LDAP_SCOPE_SUBORDINATE ? "subordinate" : "unknown"),
                 urld->lud_filter,
                 sec->parsed_url.secure == APR_LDAP_TLS_SSL  ? "using SSL": "not using SSL"
                 );

    return NULL;
}

static const char *meta_ldap_set_key(cmd_parms *cmd, void *config, const char *arg)
{
    meta_ldap_config_t *sec = config;

    const char *expr_err = NULL;
    ap_expr_info_t *expr;

    expr = ap_expr_parse_cmd(cmd, arg, AP_EXPR_FLAG_STRING_RESULT,
                             &expr_err, NULL);

    if (expr_err)
        return apr_pstrcat(cmd->temp_pool,
                           "Cannot parse expression: ",
                           expr_err, NULL);

    sec->key = expr;
    sec->key_set = 1;

    return NULL;
}

static const char *meta_ldap_set_filter(cmd_parms *cmd, void *config, const char *arg)
{            
    meta_ldap_config_t *sec = config;
            
    const char *expr_err = NULL;                    
    ap_expr_info_t *expr;
        
    expr = ap_expr_parse_cmd(cmd, arg, AP_EXPR_FLAG_STRING_RESULT,
                             &expr_err, NULL);
    
    if (expr_err)
        return apr_pstrcat(cmd->temp_pool,
                           "Cannot parse expression: ",
                           expr_err, NULL);
    
    sec->filter = expr;
    sec->filter_set = 1;

    return NULL;
}       

static const char *meta_ldap_set_deref(cmd_parms *cmd, void *config, const char *arg)
{
    meta_ldap_config_t *sec = config;

    if (strcmp(arg, "never") == 0 || strcasecmp(arg, "off") == 0) {
        sec->deref = APR_LDAP_DEREF_NEVER;
    }
    else if (strcmp(arg, "searching") == 0) {
        sec->deref = APR_LDAP_DEREF_SEARCHING;
    }
    else if (strcmp(arg, "finding") == 0) {
        sec->deref = APR_LDAP_DEREF_FINDING;
    }
    else if (strcmp(arg, "always") == 0 || strcasecmp(arg, "on") == 0) {
        sec->deref = APR_LDAP_DEREF_ALWAYS;
    }
    else {
        return "Unrecognized value for AuthLDAPDereferenceAliases directive";
    }

    sec->deref_set = 1;

    return NULL;
}

static const char *set_bind_dn(cmd_parms *cmd, void *config, const char *arg)
{
    meta_ldap_config_t *sec = config;

    sec->binddn = arg;
    sec->binddn_set = 1;

    return NULL;
}

static const char *set_bind_password(cmd_parms *cmd, void *_cfg, const char *arg)
{
    meta_ldap_config_t *sec = _cfg;
    int arglen = strlen(arg);
    char **argv;
    char *result;

    if ((arglen > 5) && strncmp(arg, "exec:", 5) == 0) {
        if (apr_tokenize_to_argv(arg+5, &argv, cmd->temp_pool) != APR_SUCCESS) {
            return apr_pstrcat(cmd->pool,
                               "Unable to parse exec arguments from ",
                               arg+5, NULL);
        }
        argv[0] = ap_server_root_relative(cmd->temp_pool, argv[0]);

        if (!argv[0]) {
            return apr_pstrcat(cmd->pool,
                               "Invalid AuthLDAPBindPassword exec location:",
                               arg+5, NULL);
        }
        result = ap_get_exec_line(cmd->pool,
                                  (const char*)argv[0], (const char * const *)argv);

        if (!result) {
            return apr_pstrcat(cmd->pool,
                               "Unable to get bind password from exec of ",
                               arg+5, NULL);
        }
        sec->bindpw = result;
    }
    else {
        sec->bindpw = (char *)arg;
    }

    if (!(*sec->bindpw)) {
        return "Empty passwords are invalid for AuthLDAPBindPassword";
    }

    sec->bindpw_set = 1;

    return NULL;
}

static const char *set_meta_prefix(cmd_parms *cmd, void *config, const char *arg)
{
    meta_ldap_config_t *sec = config;

    sec->meta_prefix = arg;
    sec->meta_prefix_set = 1;

    return NULL;
}

static const command_rec meta_ldap_cmds[] =
{
    AP_INIT_TAKE12("MetaLDAPURL", meta_ldap_parse_url, NULL, OR_AUTHCFG,
                  "URL to define LDAP connection. This should be an RFC 2255 compliant\n"
                  "URL of the form ldap://host[:port]/basedn[?attrib[?scope[?filter]]].\n"
                  "<ul>\n"
                  "<li>Host is the name of the LDAP server. Use a space separated list of hosts \n"
                  "to specify redundant servers.\n"
                  "<li>Port is optional, and specifies the port to connect to.\n"
                  "<li>basedn specifies the base DN to start searches from\n"
                  "<li>Attrib specifies what attribute to search for in the directory. If not "
                  "provided, it defaults to <b>uid</b>.\n"
                  "<li>Scope is the scope of the search, and can be either <b>sub</b> or "
                  "<b>one</b>. If not provided, the default is <b>sub</b>.\n"
                  "<li>Filter is a filter to use in the search. If not provided, "
                  "defaults to <b>(objectClass=*)</b>.\n"
                  "</ul>\n"
                  "Searches are performed using the attribute and the filter combined. "
                  "For example, assume that the\n"
                  "LDAP URL is <b>ldap://ldap.airius.com/ou=People, o=Airius?uid?sub?(posixid=*)</b>. "
                  "Searches will\n"
                  "be done using the filter <b>(&((posixid=*))(uid=<i>username</i>))</b>, "
                  "where <i>key</i>\n"
                  "is the user name pass. The search will be a subtree "
                  "search on the branch <b>ou=People, o=Airius</b>."),

    AP_INIT_TAKE1("MetaLDAPBindDN", set_bind_dn, NULL, OR_AUTHCFG,
                  "DN to use to bind to LDAP server. If not provided, will do an anonymous bind."),

    AP_INIT_TAKE1("MetaLDAPBindPassword", set_bind_password, NULL, OR_AUTHCFG,
                  "Password to use to bind to LDAP server. If not provided, will do an anonymous bind."),

    AP_INIT_TAKE1("MetaLDAPKey", meta_ldap_set_key, NULL, OR_AUTHCFG,
                  "Expression to resolve the key used for the search."),

    AP_INIT_TAKE1("MetaLDAPFilter", meta_ldap_set_filter, NULL, OR_AUTHCFG,
                  "Expression to resolve the filter used for the search. If set, overrides the "
                  "filter in the URL."),

    AP_INIT_TAKE1("MetaLDAPDereferenceAliases", meta_ldap_set_deref, NULL, OR_AUTHCFG,
                  "Determines how aliases are handled during a search. Can be one of the "
                  "values \"never\", \"searching\", \"finding\", or \"always\". "
                  "Defaults to always."),

    AP_INIT_TAKE1("MetaLDAPPrefix", set_meta_prefix, NULL, OR_AUTHCFG,
                  "The prefix to add to environment variables set during "
                  "successful lookup, default '" META_LDAP_PREFIX "'"),

    {NULL}
};

static int meta_ldap_post_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
    ap_configfile_t *f;
    char l[MAX_STRING_LEN];
    const char *charset_confname = ap_get_module_config(s->module_config,
                                                      &meta_ldap_module);
    apr_status_t status;

    /* make sure that mod_ldap (util_ldap) is loaded */
    if (ap_find_linked_module("util_ldap.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, APLOGNO(01749)
                     "Module mod_ldap missing. Mod_ldap (aka. util_ldap) "
                     "must be loaded in order for mod_meta_ldap to function properly");
        return HTTP_INTERNAL_SERVER_ERROR;

    }

    if (!charset_confname) {
        return OK;
    }

    charset_confname = ap_server_root_relative(p, charset_confname);
    if (!charset_confname) {
        ap_log_error(APLOG_MARK, APLOG_ERR, APR_EBADPATH, s, APLOGNO()
                     "Invalid charset conversion config path %s",
                     (const char *)ap_get_module_config(s->module_config,
                                                        &meta_ldap_module));
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    if ((status = ap_pcfg_openfile(&f, ptemp, charset_confname))
                != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, status, s, APLOGNO()
                     "could not open charset conversion config file %s.",
                     charset_confname);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    charset_conversions = apr_hash_make(p);

    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
        const char *ll = l;
        char *lang;

        if (l[0] == '#') {
            continue;
        }
        lang = ap_getword_conf(p, &ll);
        ap_str_tolower(lang);

        if (ll[0]) {
            char *charset = ap_getword_conf(p, &ll);
            apr_hash_set(charset_conversions, lang, APR_HASH_KEY_STRING, charset);
        }
    }
    ap_cfg_closefile(f);

    to_charset = derive_codepage_from_lang (p, "utf-8");
    if (to_charset == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, status, s, APLOGNO()
                     "could not find the UTF-8 charset in the file %s.",
                     charset_confname);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    return OK;
}

static void ImportULDAPOptFn(void)
{
    util_ldap_connection_close      = APR_RETRIEVE_OPTIONAL_FN(uldap_connection_close);
    util_ldap_connection_find_uri   = APR_RETRIEVE_OPTIONAL_FN(uldap_connection_find_uri);
    util_ldap_cache_comparedn       = APR_RETRIEVE_OPTIONAL_FN(uldap_cache_comparedn);
    util_ldap_cache_compare         = APR_RETRIEVE_OPTIONAL_FN(uldap_cache_compare);
    util_ldap_cache_checkuserid     = APR_RETRIEVE_OPTIONAL_FN(uldap_cache_checkuserid);
    util_ldap_cache_getuserdn       = APR_RETRIEVE_OPTIONAL_FN(uldap_cache_getuserdn);
    util_ldap_cache_check_subgroups = APR_RETRIEVE_OPTIONAL_FN(uldap_cache_check_subgroups);
}


static apr_status_t meta_ldap_filter(ap_filter_t *f, apr_bucket_brigade *b,
                                     ap_input_mode_t mode,
                                     apr_read_type_e block,
                                     apr_off_t readbytes)
{
    request_rec *r = f->r;
    meta_ldap_config_t *sec =
        (meta_ldap_config_t *)ap_get_module_config(r->per_dir_config, &meta_ldap_module);

    apr_array_header_t **vals;

    const char *key = NULL;
    const char *filter = NULL;
    const char *dn = NULL;
    const char *err = NULL;

    util_ldap_connection_t *ldc = NULL;

    char filtbuf[FILTER_LENGTH];

    apr_status_t status;

    ap_remove_input_filter(f);
    status = ap_get_brigade(f->next, b, mode, block, readbytes);
    if (APR_SUCCESS != status) {
        return status;
    }

    if (!sec->url_set) {
        return APR_SUCCESS;
    }

    /* Evaluate the key */
    if (sec->key) {
        key = ap_expr_str_exec(r, sec->key, &err);
        if (err) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO()
                          "meta_ldap: Can't evaluate key expression: %s",
                          err);

            apr_table_setn(f->r->subprocess_env,
                           apr_pstrcat(f->r->pool, sec->meta_prefix, "ERROR", NULL),
                           "Can't evaluate key expression");

            return APR_EINVAL;
        }
    }

    /* Evaluate the filter */
    if (sec->filter) {
        filter = ap_expr_str_exec(r, sec->filter, &err);
        if (err) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO()
                          "meta_ldap: Can't evaluate filter expression: %s",
                          err);
    
            apr_table_setn(f->r->subprocess_env,
                           apr_pstrcat(f->r->pool, sec->meta_prefix, "ERROR", NULL),
                           "Can't evaluate filter expression");

            return APR_EINVAL;
        }
    }

    /* Build the metadata filter */
    status = meta_ldap_build_filter(filtbuf, r, key, filter, sec);
    if (APR_SUCCESS != status) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO()
                "meta_ldap: ldap filter too long (>%d): %s",
                FILTER_LENGTH, filtbuf);

        apr_table_setn(f->r->subprocess_env,
                       apr_pstrcat(f->r->pool, sec->meta_prefix, "ERROR", NULL),
                       "LDAP filter too long");

        return status;
    }

    ldc = util_ldap_connection_find_uri(r, sec->url,
                                        sec->binddn, sec->bindpw,
                                        sec->deref, sec->parsed_url.secure);

    /* Search for the user DN */
    status = util_ldap_cache_getuserdn(r, ldc, sec->url, sec->parsed_url.basedn,
            sec->parsed_url.scope, sec->parsed_url.attributes, filtbuf, &dn, &(vals));

    util_ldap_connection_close(ldc);

    if (APR_STATUS_IS_NO_SUCH_OBJECT(status)) {

        /* do nothing */

        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, status, r, APLOGNO()
                "meta_ldap: search for '%s': %s", filtbuf, ldc->reason);
    }

    else if (status != APR_SUCCESS) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r, APLOGNO()
                "meta_ldap: search for '%s': %s", filtbuf, ldc->reason);

        apr_table_setn(f->r->subprocess_env,
                       apr_pstrcat(f->r->pool, sec->meta_prefix, "ERROR", NULL),
                       ldc->reason);
    }

    else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, status, r, APLOGNO()
                "meta_ldap: search for '%s': %s", filtbuf, ldc->reason);

        apr_table_setn(f->r->subprocess_env,
                       apr_pstrcat(f->r->pool, sec->meta_prefix, "DN", NULL),
                       dn);

        /* add environment variables */
        set_request_vars(f->r, vals);
    }

    return APR_SUCCESS;
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_post_config(meta_ldap_post_config,NULL,NULL,APR_HOOK_MIDDLE);

    ap_register_input_filter(META_LDAP_FILTER, meta_ldap_filter,
                             NULL, AP_FTYPE_RESOURCE);

    ap_hook_optional_fn_retrieve(ImportULDAPOptFn,NULL,NULL,APR_HOOK_MIDDLE);
}

AP_DECLARE_MODULE(meta_ldap) =
{
    STANDARD20_MODULE_STUFF,
    create_meta_ldap_dir_config,   /* dir config creater */
    merge_meta_ldap_dir_config,    /* dir merger --- default is to override */
    NULL,                          /* server config */
    NULL,                          /* merge server config */
    meta_ldap_cmds,                /* command apr_table_t */
    register_hooks                 /* register hooks */
};
